#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import numpy as np

from octobot_commons.data_util import drop_nan, mean, shift_value_array


def test_drop_nan():
    assert np.array_equal(drop_nan(np.array([1, np.nan, 2, 3, np.nan])), np.array([1, 2, 3]))
    assert np.array_equal(drop_nan(np.array([np.nan, np.nan, np.nan])), np.array([]))


def test_mean():
    assert mean([1, 2, 3, 4, 5, 6, 7]) == 4.0
    assert mean([0.684, 1, 2, 3, 4, 5.5, 6, 7.5]) == 3.7105
    assert mean([]) == 0


def test_shift_value_array():
    array = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=np.float64)
    np.testing.assert_array_equal(shift_value_array(array, shift_count=-1, fill_value=np.nan),
                                  np.array([2, 3, 4, 5, 6, 7, 8, 9, np.nan], dtype=np.float64))

    array = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9], dtype=np.float64)
    np.testing.assert_array_equal(shift_value_array(array, shift_count=2, fill_value=np.nan),
                                  np.array([np.nan, np.nan, 1, 2, 3, 4, 5, 6, 7], dtype=np.float64))



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

import octobot_commons.signals as signals
import octobot_commons.enums as enums


@pytest.fixture
def signal():
    return signals.Signal("hello topic", {"hi": "plop"})


@pytest.fixture
def signal_bundle(signal):
    return signals.SignalBundle("hello identifier", [signal], "version")


@pytest.fixture
def signal_bundle_builder():
    return signals.SignalBundleBuilder("hello builder identifier")


@pytest.fixture
def signal_builder_wrapper():
    return signals.SignalBuilderWrapper(
        "hello wrapper identifier",
        signal_builder_class=signals.SignalBundleBuilder,
        timeout=-2
    )


@pytest.fixture
def signal_dict():
    return {
        enums.SignalsAttrs.TOPIC.value:  "dict topic",
        enums.SignalsAttrs.CONTENT.value: {"dict": "content", "hi": 1},
        enums.SignalsAttrs.DEPENDENCIES.value: None,
    }


@pytest.fixture
def signal_with_dependencies_dict():
    return {
        enums.SignalsAttrs.TOPIC.value:  "dict topic",
        enums.SignalsAttrs.CONTENT.value: {"dict": "content", "hi": 1},
        enums.SignalsAttrs.DEPENDENCIES.value: {
            enums.SignalDependenciesAttrs.DEPENDENCY.value: [
                {"plop": "123"},
                {"PLIP": "456"}
            ]
        },
    }


@pytest.fixture
def signal_bundle_dict(signal_dict):
    return {
        enums.CommunityFeedAttrs.VALUE.value: {
            enums.SignalBundlesAttrs.IDENTIFIER.value: "dict identifier",
            enums.SignalBundlesAttrs.SIGNALS.value: [signal_dict],
            enums.SignalBundlesAttrs.VERSION.value: "dict_version"
        },
    }



# pylint: disable=C0103,W0603
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.constants as constants


class OptimizationCampaign:
    def __init__(self, name=None):
        self.name = name or self.get_campaign_name()

    @classmethod
    def get_campaign_name(cls, *args):
        """
        Returns the name of the current optimization campaign
        :param args: arguments passed to the optimization_campaign_name_proxy
        """
        return _optimization_name_proxy(*args)


def _default_optimization_name_proxy(*_):
    return constants.DEFAULT_CAMPAIGN


_name_proxy = _default_optimization_name_proxy


def _optimization_name_proxy(*args):
    return _name_proxy(*args)


def register_optimization_campaign_name_proxy(new_proxy):
    """
    Registers a new campaign name provider as a proxy function
    :param new_proxy: the proxy function to be called by OptimizationCampaign.get_campaign_name
    """
    global _name_proxy
    _name_proxy = new_proxy



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time
import mock
from datetime import timezone, timedelta

from octobot_commons.timestamp_util import is_valid_timestamp, get_now_time, datetime_to_timestamp, \
    convert_timestamp_to_datetime


def test_is_valid_timestamp():
    assert not is_valid_timestamp(get_now_time())
    assert is_valid_timestamp(time.time())


def test_datetime_to_timestamp():
    date_str = convert_timestamp_to_datetime(1737331200, time_format="%d/%m/%y %H:%M", local_timezone=True)
    assert datetime_to_timestamp(date_str, "%d/%m/%y %H:%M") == 1737331200


def test_convert_timestamp_to_datetime_default():
    # 1 Jan 2020 00:00 UTC
    ts = 1577836800
    assert convert_timestamp_to_datetime(ts) == "01/01/20 00:00"


def test_convert_timestamp_to_datetime_custom_format():
    ts = 1577836800
    assert convert_timestamp_to_datetime(ts, time_format="%Y-%m-%d") == "2020-01-01"


def test_convert_timestamp_to_datetime_local_timezone(monkeypatch):
    # Patch LOCAL_TIMEZONE to UTC+2 for test
    dummy_tz = timezone(timedelta(hours=2))
    with mock.patch("octobot_commons.timestamp_util.LOCAL_TIMEZONE", dummy_tz):
        ts = 1577836800  # 1 Jan 2020 00:00 UTC
        # Should be 02:00 in UTC+2
        assert convert_timestamp_to_datetime(ts, local_timezone=True) == "01/01/20 02:00"


def test_convert_timestamp_to_datetime_epoch():
    assert convert_timestamp_to_datetime(0) == "01/01/70 00:00"


def test_convert_timestamp_to_datetime_far_future():
    ts = 32503680000  # year 3000
    assert convert_timestamp_to_datetime(ts, time_format="%Y") == "3000"


def test_convert_timestamp_to_datetime_negative():
    # Negative timestamp: before epoch
    ts = -1
    result = convert_timestamp_to_datetime(ts)
    # Should not raise, but result is system-dependent
    assert isinstance(result, str)



# pylint: disable=C0116
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import time

import octobot_commons.signals.signal_bundle_builder as signal_bundle_builder


class SignalBuilderWrapper:
    NO_TIMEOUT_VALUE = -1

    def __init__(
        self,
        identifier: str,
        signal_builder_class=signal_bundle_builder.SignalBundleBuilder,
        timeout: float = NO_TIMEOUT_VALUE,
        builder_args: tuple = None,
    ):
        self.signal_builder_class = signal_builder_class
        self.signal_bundle_builder = (
            signal_builder_class(identifier, *builder_args)
            if builder_args
            else signal_builder_class(identifier)
        )
        self.is_being_emitted = False
        self.timeout = timeout
        self.timeout_event = asyncio.Event()
        self.signal_emit_time = time.time() + timeout
        self._users_count = 0

    def register_user(self):
        self._users_count += 1

    def unregister_user(self):
        self._users_count -= 1

    def has_single_user(self):
        return self._users_count == 1



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.dsl_interpreter.operator as dsl_interpreter_operator


class SubscriptingOperator(
    dsl_interpreter_operator.Operator
):  # pylint: disable=abstract-method
    """
    Base class for subscripting operators (array/list subscripting).
    Subscripting operators have 3 parameters: the array/list, the index or slice and the context.
    # https://docs.python.org/3/library/ast.html#subscripting
    """



# pylint: disable=C0116
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing

import octobot_commons.enums


class SignalDependencies:
    def __init__(self, dependencies: typing.Optional[list[dict]] = None):
        self.dependencies: list[dict] = dependencies if dependencies else []

    def extend(self, dependencies: "SignalDependencies"):
        self.dependencies.extend(dependencies.dependencies)

    def is_filled_by(self, filled_dependencies: "SignalDependencies") -> bool:
        for dependency in self.dependencies:
            has_missing_dependency = True
            for filled_dependency in filled_dependencies.dependencies:
                has_missing_dependency = bool(
                    {
                        key: value
                        for key, value in dependency.items()
                        if key not in filled_dependency
                        or filled_dependency[key] != value
                    }
                )
                if not has_missing_dependency:
                    break
            # iterated over all filled_dependency and did not break,
            # this dependency is not filled
            if has_missing_dependency:
                return False
        # all dependencies are filled
        return True

    def to_dict(self) -> dict:
        return {
            octobot_commons.enums.SignalDependenciesAttrs.DEPENDENCY.value: self.dependencies,
        }

    def __str__(self):
        return f"{self.to_dict()}"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other: "SignalDependencies") -> bool:
        if self is other:
            return True
        if not isinstance(other, SignalDependencies):
            return False
        return self.dependencies == other.dependencies

    def __bool__(self) -> bool:
        return bool(self.dependencies)



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import octobot_commons.constants as constants
import octobot_commons.configuration as configuration


class TestHistoricalConfiguration:
    """Test class for historical configuration functions"""

    def setup_method(self):
        """Set up test fixtures"""
        self.empty_master_config = {}
        self.master_config_with_historical = {} 
        for ts, config in (
          [1000.0, {"param1": "value1", "param2": "value2"}],
          [2000.0, {"param1": "value3", "param2": "value4"}],
          [3000.0, {"param1": "value5", "param2": "value6"}],
        ):
          # add config using add_historical_tentacle_config to maintain order
          configuration.add_historical_tentacle_config(self.master_config_with_historical, ts, config)
        
        self.test_config = {"test_param": "test_value"}
        self.test_start_time = 1500.0

    def test_add_historical_tentacle_config_empty_master_config(self):
        assert constants.CONFIG_HISTORICAL_CONFIGURATION not in self.empty_master_config
        """Test adding historical config to empty master config"""
        configuration.add_historical_tentacle_config(
            self.empty_master_config, self.test_start_time, self.test_config
        )
        
        assert constants.CONFIG_HISTORICAL_CONFIGURATION in self.empty_master_config
        assert len(self.empty_master_config[constants.CONFIG_HISTORICAL_CONFIGURATION]) == 1
        assert self.empty_master_config[constants.CONFIG_HISTORICAL_CONFIGURATION][0] == [
            self.test_start_time, self.test_config
        ]

    def test_add_historical_tentacle_config_existing_master_config(self):
        """Test adding historical config to existing master config"""
        original_length = len(self.master_config_with_historical[constants.CONFIG_HISTORICAL_CONFIGURATION])
        
        configuration.add_historical_tentacle_config(
            self.master_config_with_historical, self.test_start_time, self.test_config
        )
        
        # Check that the new config was added
        assert len(self.master_config_with_historical[constants.CONFIG_HISTORICAL_CONFIGURATION]) == original_length + 1
        
        # Check that the list is sorted by timestamp in descending order (most recent first)
        timestamps = [config[0] for config in self.master_config_with_historical[constants.CONFIG_HISTORICAL_CONFIGURATION]]
        assert timestamps == sorted(timestamps, reverse=True)
        
        # Check that our new config is in the list
        config_found = False
        for config in self.master_config_with_historical[constants.CONFIG_HISTORICAL_CONFIGURATION]:
            if config[0] == self.test_start_time and config[1] == self.test_config:
                config_found = True
                break
        assert config_found

    def test_add_historical_tentacle_config_multiple_additions(self):
        """Test adding multiple historical configs maintains sorting"""
        configuration.add_historical_tentacle_config(self.empty_master_config, 1000.0, {"config": "old"})
        configuration.add_historical_tentacle_config(self.empty_master_config, 3000.0, {"config": "new"})
        configuration.add_historical_tentacle_config(self.empty_master_config, 2000.0, {"config": "middle"})
        
        timestamps = [config[0] for config in self.empty_master_config[constants.CONFIG_HISTORICAL_CONFIGURATION]]
        assert timestamps == [3000.0, 2000.0, 1000.0]

    def test_has_any_historical_tentacle_config_empty(self):
        """Test has_any_historical_tentacle_config with empty config"""
        assert not configuration.has_any_historical_tentacle_config(self.empty_master_config)

    def test_has_any_historical_tentacle_config_with_historical(self):
        """Test has_any_historical_tentacle_config with existing historical config"""
        assert configuration.has_any_historical_tentacle_config(self.master_config_with_historical)

    def test_has_any_historical_tentacle_config_other_keys(self):
        """Test has_any_historical_tentacle_config with other keys but not historical"""
        config_with_other_keys = {"other_key": "value", "another_key": "value2"}
        assert not configuration.has_any_historical_tentacle_config(config_with_other_keys)

    def test_get_historical_tentacle_config_exact_match(self):
        """Test getting historical config with exact time match"""
        result = configuration.get_historical_tentacle_config(self.master_config_with_historical, 2000.0)
        assert result == {"param1": "value3", "param2": "value4"}

    def test_get_historical_tentacle_config_between_times(self):
        """Test getting historical config with time between existing configs"""
        result = configuration.get_historical_tentacle_config(self.master_config_with_historical, 1500.0)
        assert result == {"param1": "value1", "param2": "value2"}

    def test_get_historical_tentacle_config_before_oldest(self):
        """Test getting historical config with time before oldest config"""
        result = configuration.get_historical_tentacle_config(self.master_config_with_historical, 500.0)
        # Should return the oldest config (last in the list)
        assert result == {"param1": "value1", "param2": "value2"}

    def test_get_historical_tentacle_config_after_newest(self):
        """Test getting historical config with time after newest config"""
        result = configuration.get_historical_tentacle_config(self.master_config_with_historical, 4000.0)
        # Should return the newest config (first in the list)
        assert result == {"param1": "value5", "param2": "value6"}

    def test_get_historical_tentacle_config_key_error(self):
        """Test get_historical_tentacle_config raises KeyError when key not found"""
        with pytest.raises(KeyError) as exc_info:
            configuration.get_historical_tentacle_config(self.empty_master_config, 1000.0)
        assert constants.CONFIG_HISTORICAL_CONFIGURATION in str(exc_info.value)

    def test_get_historical_tentacle_configs_empty_interval(self):
        """Test getting historical configs with empty time interval"""
        result = configuration.get_historical_tentacle_configs(self.master_config_with_historical, 5000.0, 6000.0)
        assert result == []

    def test_get_historical_tentacle_configs_full_interval(self):
        """Test getting historical configs with full time interval"""
        result = configuration.get_historical_tentacle_configs(self.master_config_with_historical, 500.0, 4000.0)
        assert len(result) == 3
        # Should be ordered by most recent first
        assert result[0] == {"param1": "value5", "param2": "value6"}
        assert result[1] == {"param1": "value3", "param2": "value4"}
        assert result[2] == {"param1": "value1", "param2": "value2"}

    def test_get_historical_tentacle_configs_partial_interval(self):
        """Test getting historical configs with partial time interval"""
        result = configuration.get_historical_tentacle_configs(self.master_config_with_historical, 1500.0, 2500.0)
        assert len(result) == 1
        assert result[0] == {"param1": "value3", "param2": "value4"}

    def test_get_historical_tentacle_configs_exact_boundaries(self):
        """Test getting historical configs with exact boundary times"""
        result = configuration.get_historical_tentacle_configs(self.master_config_with_historical, 1000.0, 3000.0)
        assert len(result) == 3

    def test_get_historical_tentacle_configs_key_error(self):
        """Test get_historical_tentacle_configs raises KeyError when key not found"""
        with pytest.raises(KeyError) as exc_info:
            configuration.get_historical_tentacle_configs(self.empty_master_config, 1000.0, 2000.0)
        assert constants.CONFIG_HISTORICAL_CONFIGURATION in str(exc_info.value)

    def test_get_oldest_historical_tentacle_config_time_success(self):
        """Test getting oldest historical config time successfully"""
        result = configuration.get_oldest_historical_tentacle_config_time(self.master_config_with_historical)
        assert result == 1000.0

    def test_get_oldest_historical_tentacle_config_time_empty_list(self):
        """Test getting oldest historical config time with empty list"""
        config_with_empty_list = {constants.CONFIG_HISTORICAL_CONFIGURATION: []}
        with pytest.raises(ValueError) as exc_info:
            configuration.get_oldest_historical_tentacle_config_time(config_with_empty_list)
        assert "No historical configuration found" in str(exc_info.value)

    def test_get_oldest_historical_tentacle_config_time_missing_key(self):
        """Test getting oldest historical config time with missing key"""
        # Should return the minimum of an empty list, which raises ValueError
        with pytest.raises(ValueError) as exc_info:
            configuration.get_oldest_historical_tentacle_config_time(self.empty_master_config)
        assert "No historical configuration found" in str(exc_info.value)

    def test_get_oldest_historical_tentacle_config_time_single_config(self):
        """Test getting oldest historical config time with single config"""
        single_config = {
            constants.CONFIG_HISTORICAL_CONFIGURATION: [[1500.0, {"test": "config"}]]
        }
        result = configuration.get_oldest_historical_tentacle_config_time(single_config)
        assert result == 1500.0

    def test_integration_scenario(self):
        """Test integration scenario with multiple operations"""
        # Start with empty config
        master_config = {}
        
        # Add multiple configs
        configuration.add_historical_tentacle_config(master_config, 1000.0, {"version": "1.0"})
        configuration.add_historical_tentacle_config(master_config, 2000.0, {"version": "2.0"})
        configuration.add_historical_tentacle_config(master_config, 1500.0, {"version": "1.5"})
        
        # Check that historical config exists
        assert configuration.has_any_historical_tentacle_config(master_config)
        
        # Get config for different times
        assert configuration.get_historical_tentacle_config(master_config, 500.0) == {"version": "1.0"}
        assert configuration.get_historical_tentacle_config(master_config, 1200.0) == {"version": "1.0"}
        assert configuration.get_historical_tentacle_config(master_config, 1600.0) == {"version": "1.5"}
        assert configuration.get_historical_tentacle_config(master_config, 2500.0) == {"version": "2.0"}
        
        # Get configs for time intervals
        configs_1000_2000 = configuration.get_historical_tentacle_configs(master_config, 1000.0, 2000.0)
        assert len(configs_1000_2000) == 3
        
        # Get oldest time
        assert configuration.get_oldest_historical_tentacle_config_time(master_config) == 1000.0



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import pytest

import octobot_commons.asyncio_tools as asyncio_tools

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_without_error_container():
    # will not propagate exception
    asyncio.get_event_loop().call_soon(_exception_raiser)


async def test_with_error_container():
    error_container = asyncio_tools.ErrorContainer()
    error_container.print_received_exceptions = False
    asyncio.get_event_loop().set_exception_handler(error_container.exception_handler)
    # will propagate exception
    asyncio.get_event_loop().call_soon(_exception_raiser)
    with pytest.raises(AssertionError):
        # ensure exception is caught
        await asyncio.create_task(error_container.check())


async def test_with_error_container_2_exceptions():
    error_container = asyncio_tools.ErrorContainer()
    error_container.print_received_exceptions = False
    asyncio.get_event_loop().set_exception_handler(error_container.exception_handler)
    # will propagate exception
    asyncio.get_event_loop().call_soon(_exception_raiser)
    asyncio.get_event_loop().call_soon(_exception_raiser)
    with pytest.raises(AssertionError):
        # ensure exception is caught
        await asyncio.create_task(error_container.check())


async def test_RLock_valid_setup():
    lock_1 = asyncio_tools.RLock()
    lock_2 = asyncio_tools.RLock()

    passed_a = passed_b = passed_c = passed_d = passed_e = False
    async with lock_1:
        passed_a = True
        async with lock_1:
            passed_b = True
            assert lock_1._depth == 2
            async with lock_2:
                passed_c = True
                async with lock_2:
                    assert lock_2._depth == 2
                    passed_d = True
                    async with lock_1:
                        assert lock_1._task is asyncio.current_task()
                        assert lock_2._task is asyncio.current_task()
                        assert lock_1._depth == 3
                        passed_e = True
                    assert lock_1._depth == 2
    assert lock_1._depth == 0
    assert lock_2._depth == 0
    assert lock_1._locked is False
    assert lock_2._locked is False
    assert lock_1._task is None
    assert lock_2._task is None
    assert all((passed_a, passed_b, passed_c, passed_d, passed_e))


async def test_RLock_multiple_tasks():
    lock = asyncio_tools.RLock()
    started = {
        "a": False,
        "b": False,
        "c": False,
    }
    passed = {
        "a": False,
        "b": False,
        "c": False,
    }
    released = {
        "a": False,
        "b": False,
        "c": False,
    }

    async def passing_task(identifier, is_waiting):
        started[identifier] = True
        async with lock:
            if is_waiting:
                await asyncio_tools.wait_asyncio_next_cycle()
            assert lock._depth == 1
            assert lock._task == asyncio.current_task()
            passed[identifier] = True
        released[identifier] = True

    asyncio.create_task(passing_task("a", True))
    asyncio.create_task(passing_task("b", False))
    asyncio.create_task(passing_task("c", False))

    # tasks did not run yet
    assert all(v is False for v in started.values())

    # let the loop run once
    await asyncio_tools.wait_asyncio_next_cycle()

    # all tasks started
    assert started == {
        "a": True,
        "b": True,
        "c": True,
    }
    # "a" waits for next cycle, others wait for lock that "a" got
    assert passed == {
        "a": False,
        "b": False,
        "c": False,
    }
    assert released == {
        "a": False,
        "b": False,
        "c": False,
    }
    await asyncio_tools.wait_asyncio_next_cycle()
    # "a" release the lock, "b" can get it
    assert passed == {
        "a": True,
        "b": True,
        "c": False,
    }
    assert released == {
        "a": True,
        "b": True,
        "c": False,
    }
    # "b" release the lock, "c" can get it
    await asyncio_tools.wait_asyncio_next_cycle()

    assert all(v is True for v in passed.values())
    assert all(v is True for v in released.values())


async def test_RLock_error_setup_1():
    lock = asyncio_tools.RLock()

    passed_a = passed_b = False

    # _depth is too high: lock is not cleared
    async with lock:
        passed_a = True
        lock._depth += 1
        async with lock:
            passed_b = True
            assert lock._depth == 3
    assert lock._depth == 1
    assert lock._task is asyncio.current_task()
    assert lock._locked is True

    assert all((passed_a, passed_b))


async def test_RLock_error_setup_2():
    lock = asyncio_tools.RLock()
    lock._depth = 1

    # a lock can't have a non 0 depth when firstly acquired
    with pytest.raises(RuntimeError):
        async with lock:
            pass


def _exception_raiser():
    raise RuntimeError("error")



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import sys

import octobot_commons.singleton as singleton


class GlobalSharedMemoryStorage(dict, singleton.Singleton):
    """
    A global singleton dict available to the whole python virtual machine.
    Warnings:
        only stored in RAM, not persisted on disc
        not thread safe
    """

    def remove_oldest_elements(self, elements_count_to_remove: int):
        """
        Remove (pop) the elements_count_to_remove oldest elements
        :param elements_count_to_remove: number of elements to remove
        """
        for key in list(self.keys())[:elements_count_to_remove]:
            self.pop(key)

    def get_bytes_size(self):
        """
        Return the size in bytes of the memory storage
        """
        return sys.getsizeof(self)



# pylint: disable=too-many-branches,too-many-return-statements,too-many-locals,too-many-statements
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ast
import typing
import octobot_commons.errors
import octobot_commons.dsl_interpreter.operator as dsl_interpreter_operator
import octobot_commons.dsl_interpreter.interpreter_dependency as dsl_interpreter_dependency


class Interpreter:
    """
    Interpreter class for parsing and interpreting DSL expressions.
    Maps AST elements to Operator instances based on operator names.
    """

    def __init__(
        self, operators: typing.List[typing.Type[dsl_interpreter_operator.Operator]]
    ):
        """
        Initialize the interpreter with a list of operator classes.

        Args:
            operators: List of Operator subclasses to be used for interpretation
        """
        # Save operators as a dictionary mapping operator name to operator class
        self.operators_by_name: typing.Dict[
            str, typing.Type[dsl_interpreter_operator.Operator]
        ] = {}
        self.extend(operators)
        self._operator_tree_or_constant: typing.Union[
            dsl_interpreter_operator.Operator,
            dsl_interpreter_operator.ComputedOperatorParameterType,
        ] = None

    def extend(
        self, operators: typing.List[typing.Type[dsl_interpreter_operator.Operator]]
    ):
        """
        Extend the interpreter with a list of operator classes.
        """
        self.operators_by_name.update(
            {operator_class.get_name(): operator_class for operator_class in operators}
        )

    async def interprete(
        self, expression: str
    ) -> dsl_interpreter_operator.ComputedOperatorParameterType:
        """
        Interpret a string expression by parsing it with AST and mapping to operators.

        Args:
            expression: String expression to interpret

        Returns:
            Operator instance or literal value representing the interpreted expression
        """
        self._parse_expression(expression)
        return await self.compute_expression()

    def get_dependencies(
        self,
    ) -> typing.List[dsl_interpreter_dependency.InterpreterDependency]:
        """
        Get the dependencies of the interpreter's parsed expression.
        """
        if self._operator_tree_or_constant is None:
            raise octobot_commons.errors.DSLInterpreterError(
                "Expression not prepared, call prepare() first"
            )
        if isinstance(
            self._operator_tree_or_constant, dsl_interpreter_operator.Operator
        ):
            all_deps = self._operator_tree_or_constant.get_dependencies()
            # don't use set to avoid forcing the dependency to implement __hash__
            deduplicated_deps = []
            for dep in all_deps:
                if dep not in deduplicated_deps:
                    deduplicated_deps.append(dep)
            return deduplicated_deps
        return []

    def prepare(self, expression: str):
        """
        Prepare the expression by parsing it, leaving it ready to be computed one or multiple times.
        Call await self.compute_expression() to compute the expression.
        """
        self._parse_expression(expression)

    def _parse_expression(self, expression: str):
        """
        Parse the expression into an AST and store the result in self._operator_tree_or_constant.
        """
        # Parse the expression into an AST
        # mode:  can be 'exec' if source consists of a sequence of statements, 'eval' if
        # it consists of a single expression, or 'single' if it consists of a single
        # interactive statement.
        # docs: https://docs.python.org/3/library/functions.html#compile
        tree = ast.parse(expression, mode="eval")

        # Visit the AST and convert nodes to Operator instances
        self._operator_tree_or_constant = self._visit_node(tree.body)

    async def compute_expression(
        self,
    ) -> dsl_interpreter_operator.ComputedOperatorParameterType:
        """
        Compute the result of the expression stored in self._operator_tree_or_constant.
        If the expression is a constant, return it directly.
        If the expression is an operator, pre_compute and compute its result.
        """
        if isinstance(
            self._operator_tree_or_constant, dsl_interpreter_operator.Operator
        ):
            await self._operator_tree_or_constant.pre_compute()
            return self._operator_tree_or_constant.compute()
        return self._operator_tree_or_constant

    def _visit_node(
        self, node: typing.Optional[ast.AST]
    ) -> typing.Union[
        dsl_interpreter_operator.Operator,
        dsl_interpreter_operator.ComputedOperatorParameterType,
    ]:
        """
        Recursively visit AST nodes and convert them to Operator instances or values.

        Args:
            node: AST node to visit

        Returns:
            Operator instance or literal value representing the node
        """
        if node is None:
            return None

        if isinstance(node, ast.Call):
            # Function call: func(arg1, arg2, ...)
            func_name = self._get_name_from_node(node.func)
            if func_name in self.operators_by_name:
                operator_class = self.operators_by_name[func_name]
                # Convert arguments to Operator instances or values
                args = [
                    self._get_value_from_constant_node(arg)
                    if isinstance(arg, ast.Constant)
                    else self._visit_node(arg)
                    for arg in node.args
                ]
                return operator_class(*args)
            raise octobot_commons.errors.UnsupportedOperatorError(
                f"Unknown operator: {func_name}"
            )

        if isinstance(node, ast.BinOp):
            # Binary operation: left op right
            op_name = type(node.op).__name__
            if op_name in self.operators_by_name:
                operator_class = self.operators_by_name[op_name]
                left = self._visit_node(node.left)
                right = self._visit_node(node.right)
                return operator_class(left, right)
            raise octobot_commons.errors.UnsupportedOperatorError(
                f"Unknown binary operator: {op_name}"
            )

        if isinstance(node, ast.UnaryOp):
            # Unary operation: op operand
            op_name = type(node.op).__name__
            if op_name in self.operators_by_name:
                operator_class = self.operators_by_name[op_name]
                operand = self._visit_node(node.operand)
                return operator_class(operand)
            raise octobot_commons.errors.UnsupportedOperatorError(
                f"Unknown unary operator: {op_name}"
            )

        if isinstance(node, ast.Compare):
            # Comparison: left op right
            if len(node.ops) == 1 and len(node.comparators) == 1:
                op_name = type(node.ops[0]).__name__
                if op_name in self.operators_by_name:
                    operator_class = self.operators_by_name[op_name]
                    left = self._visit_node(node.left)
                    right = self._visit_node(node.comparators[0])
                    return operator_class(left, right)
                raise octobot_commons.errors.UnsupportedOperatorError(
                    f"Unknown comparison operator: {op_name}"
                )
            raise octobot_commons.errors.UnsupportedOperatorError(
                "Multiple comparisons not supported"
            )

        if isinstance(node, (ast.Constant)):
            # Literal values: numbers, strings, booleans, None
            return self._get_value_from_constant_node(node)

        if isinstance(node, ast.Name):
            # Name reference: look up in operators_by_name
            name = node.id
            if name in self.operators_by_name:
                operator_class = self.operators_by_name[name]
                return operator_class()
            raise octobot_commons.errors.UnsupportedOperatorError(
                f"Unknown name: {name}"
            )

        if isinstance(node, ast.BoolOp):
            # Boolean operation: left op right
            op_name = type(node.op).__name__
            if op_name in self.operators_by_name:
                operator_class = self.operators_by_name[op_name]
                operands = [self._visit_node(operand) for operand in node.values]
                return operator_class(*operands)
            raise octobot_commons.errors.UnsupportedOperatorError(
                f"Unknown BoolOp operator: {op_name}"
            )

        if isinstance(node, ast.IfExp):
            # If expression: "body if test else orelse"
            op_name = ast.IfExp.__name__
            if op_name in self.operators_by_name:
                operator_class = self.operators_by_name[op_name]
                test = self._visit_node(node.test)
                body = self._visit_node(node.body)
                orelse = self._visit_node(node.orelse)
                return operator_class(test, body, orelse)
            raise octobot_commons.errors.UnsupportedOperatorError(
                f"Unknown IfExp operator: {op_name}"
            )

        if isinstance(node, ast.Subscript):
            # Subscript: array[index]
            op_name = type(node).__name__
            if op_name in self.operators_by_name:
                operator_class = self.operators_by_name[op_name]
                array_or_list = self._visit_node(node.value)
                index_or_slice = self._visit_node(node.slice)
                context = node.ctx
                return operator_class(array_or_list, index_or_slice, context)

        if isinstance(node, ast.List):
            # List: [1, 2, 3]
            op_name = ast.List.__name__
            if op_name in self.operators_by_name:
                operator_class = self.operators_by_name[op_name]
                operands = [self._visit_node(operand) for operand in node.elts]
                return operator_class(*operands)

        if isinstance(node, ast.Slice):
            # Slice: slice(1, 2, 3)
            op_name = ast.Slice.__name__
            if op_name in self.operators_by_name:
                operator_class = self.operators_by_name[op_name]
                lower = self._visit_node(node.lower)
                upper = self._visit_node(node.upper)
                step = self._visit_node(node.step)
                return operator_class(lower, upper, step)

        raise octobot_commons.errors.UnsupportedOperatorError(
            f"Unsupported AST node type: {type(node).__name__}"
        )

    def _get_name_from_node(self, node: ast.AST) -> str:
        """Extract the name from a function node."""
        if isinstance(node, ast.Name):
            return node.id
        # elif isinstance(node, ast.Attribute): ex: snake.colour => not supported
        #     return node.attr
        raise octobot_commons.errors.UnsupportedOperatorError(
            f"Cannot extract name from node type: {type(node).__name__}"
        )

    def _get_value_from_constant_node(
        self, node: ast.Constant
    ) -> dsl_interpreter_operator.ComputedOperatorParameterType:
        """Extract a literal value from an AST constant node."""
        value = node.value
        # Filter out unsupported types like complex numbers or Ellipsis
        if isinstance(value, (str, int, float, bool, type(None))):
            return value
        raise octobot_commons.errors.UnsupportedOperatorError(
            f"Unsupported constant type: {type(value).__name__}"
        )



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums
import octobot_commons.signals as signals


from tests.signals import signal


def test_to_dict(signal):
    assert signal.to_dict() == {
        octobot_commons.enums.SignalsAttrs.TOPIC.value: "hello topic",
        octobot_commons.enums.SignalsAttrs.CONTENT.value: {"hi": "plop"},
        octobot_commons.enums.SignalsAttrs.DEPENDENCIES.value: None,
    }
    signal.dependencies = signals.SignalDependencies([
        {"plop": "123"},
        {"PLIP": "123"}
    ])
    assert signal.to_dict() == {
        octobot_commons.enums.SignalsAttrs.TOPIC.value: "hello topic",
        octobot_commons.enums.SignalsAttrs.CONTENT.value: {"hi": "plop"},
        octobot_commons.enums.SignalsAttrs.DEPENDENCIES.value: {
            octobot_commons.enums.SignalDependenciesAttrs.DEPENDENCY.value: [
                {"plop": "123"},
                {"PLIP": "123"}
            ]
        },
    }


def test__str__(signal):
    assert all(sub_str in str(signal) for sub_str in ("hello topic", "hi", "plop"))



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.dict_util as dict_util


def test_find_nested_value():
    assert dict_util.find_nested_value({"a": 1, "b": 2}, "b") == (True, 2)
    assert dict_util.find_nested_value({"a": 1, "b": 2}, "c") == (False, "c")
    assert dict_util.find_nested_value({"a": 1, "b": {"c": 5, "d": "e"}}, "d") == (True, "e")
    assert dict_util.find_nested_value({"a": 1, "b": {"c": 5, "d": {"f": [1, 2, 3]}}}, "f") == (True, [1, 2, 3])
    assert dict_util.find_nested_value({"a": {"e": 7}, "b": {"c": {"t": 5}, "d": {"f": [1, 2, 3], "y": 1}}}, "y") == (True, 1)
    complex_dict = {"a": {"e": 7}, "b": {"c": {"t": [{
        "ab": [8, 9, 10],
        "cd": {
            "4": 4,
            "5": 5,
            "6": [7, 8, 9, {
                "abc": "def",
                "zyx": "zxv"
            }]
        }
    },
        {
            "up": "pa",
            "123": 1234
        }]}, "d": {"f": [1, 2, 3], "y": 1}}}
    assert dict_util.find_nested_value(complex_dict, "abc") == (True, "def")
    assert dict_util.find_nested_value(complex_dict, "5") == (True, 5)
    assert dict_util.find_nested_value(complex_dict, "123") == (True, 1234)


def test_check_and_merge_values_from_reference():
    current_dict = {
        "b": 5,
        "d": 1
    }
    ref_dict = {
        "a": 1,
        "b": 2,
        "c": 3,
        "d": 4
    }
    exception_list = ["d"]
    dict_util.check_and_merge_values_from_reference(current_dict, ref_dict, exception_list)
    assert current_dict == {
        "a": 1,
        "b": 5,
        "c": 3,
        "d": 1
    }


def test_contains_each_element():
    assert dict_util.contains_each_element(
        {},
        {}
    ) is True
    assert dict_util.contains_each_element(
        {},
        {"1": 1}
    ) is False
    assert dict_util.contains_each_element(
        {"1": 1},
        {"1": 1}
    ) is True
    assert dict_util.contains_each_element(
        {"1": 1, "2": 2},
        {"1": 1}
    ) is True
    assert dict_util.contains_each_element(
        {"1": 1},
        {"1": 1, "2": 2}
    ) is False
    assert dict_util.contains_each_element(
        {"1": 1},
        {}
    ) is True
    assert dict_util.contains_each_element(
        {"1": 2},
        {"1": 1}
    ) is False


def test_nested_update_dict():
    assert dict_util.nested_update_dict({}, {"1": 1, "2": 2}) == {"1": 1, "2": 2}
    assert dict_util.nested_update_dict({"1": "aa"}, {"1": 1, "2": 2}) == {"1": 1, "2": 2}
    assert dict_util.nested_update_dict({"1": []}, {"1": 1, "2": 2}) == {"1": 1, "2": 2}
    assert dict_util.nested_update_dict(
        {"1": []}, {"1": 1, "2": 2}, ignore_lists=True
    ) == {"1": [], "2": 2}
    assert dict_util.nested_update_dict(
        {"1": [], "plop": {"jerome": 0, "michel": 1, "monique": [1]}},
        {"1": 1, "2": 2,  "plop": {"jerome": 0.5, "simon": 3, "monique": [2]}},
        ignore_lists=True
    ) == {"1": [], "2": 2, "plop": {"jerome": 0.5, "michel": 1, "monique": [1], "simon": 3}}
    assert dict_util.nested_update_dict(
        {"1": [], "plop": {"jerome": 0, "michel": 1, "monique": [1]}},
        {"1": 1, "2": 2,  "plop": {"jerome": 0.5, "simon": 3, "monique": [2]}},
        ignore_lists=False
    ) == {"1": 1, "2": 2, "plop": {"jerome": 0.5, "michel": 1, "monique": [2], "simon": 3}}
    assert dict_util.nested_update_dict(
        {"1": []},
        {"1": [{"jerome": 0.5, "monique": [2]}, {"simon": 3, "monique": [23]}]},
        ignore_lists=True
    ) == {"1": []}
    assert dict_util.nested_update_dict(
        {"1": []},
        {"1": [{"jerome": 0.5, "monique": [2]}, {"simon": 3, "monique": [23]}]},
        ignore_lists=False
    ) == {'1': [{'jerome': 0.5, 'monique': [2]}, {'monique': [23], 'simon': 3}]}



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import enum


class OctoBotChannelsName(enum.Enum):
    """
    OctoBot-Evaluators channel names
    """

    OCTOBOT_CHANNEL = "OctoBot"


class OctoBotUserChannelsName(enum.Enum):
    """
    OctoBot-Backtesting channel names
    """

    USER_COMMANDS_CHANNEL = "UserCommands"


class OctoBotEvaluatorsChannelsName(enum.Enum):
    """
    OctoBot-Evaluators channel names
    """

    MATRIX_CHANNEL = "Matrix"
    EVALUATORS_CHANNEL = "Evaluators"


class OctoBotBacktestingChannelsName(enum.Enum):
    """
    OctoBot-Backtesting channel names
    """

    TIME_CHANNEL = "Time"


class OctoBotCommunityChannelsName(enum.Enum):
    """
    OctoBot community channel names
    """

    REMOTE_TRADING_SIGNALS_CHANNEL = "RemoteTradingSignals"


class OctoBotTradingChannelsName(enum.Enum):
    """
    OctoBot-Trading channel names
    """

    OHLCV_CHANNEL = "OHLCV"
    TICKER_CHANNEL = "Ticker"
    MINI_TICKER_CHANNEL = "MiniTicker"
    RECENT_TRADES_CHANNEL = "RecentTrade"
    ORDER_BOOK_CHANNEL = "OrderBook"
    ORDER_BOOK_TICKER_CHANNEL = "OrderBookTicker"
    KLINE_CHANNEL = "Kline"
    TRADES_CHANNEL = "Trades"
    LIQUIDATIONS_CHANNEL = "Liquidations"
    ORDERS_CHANNEL = "Orders"
    BALANCE_CHANNEL = "Balance"
    BALANCE_PROFITABILITY_CHANNEL = "BalanceProfitability"
    POSITIONS_CHANNEL = "Positions"
    MODE_CHANNEL = "Mode"
    MARK_PRICE_CHANNEL = "MarkPrice"
    FUNDING_CHANNEL = "Funding"



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums
import octobot_commons.signals as signals

from tests.signals import signal_bundle_builder


def test_register_signal(signal_bundle_builder):
    assert signal_bundle_builder.signals == []
    signal_bundle_builder.register_signal("plop_topic", {"content_key1": 0}, other_kwarg=11)
    assert len(signal_bundle_builder.signals) == 1
    assert signal_bundle_builder.signals[0].topic == "plop_topic"
    assert signal_bundle_builder.signals[0].content == {"content_key1": 0}


def test_create_signal(signal_bundle_builder):
    created_signal = signal_bundle_builder.create_signal("plop_topic", {"content_key1": 0}, other_kwarg=11)
    assert created_signal.topic == "plop_topic"
    assert created_signal.content == {"content_key1": 0}


def test_is_empty(signal_bundle_builder):
    assert signal_bundle_builder.is_empty()
    signal_bundle_builder.register_signal("plop_topic", {"content_key1": 0}, other_kwarg=11)
    assert not signal_bundle_builder.is_empty()


def test_build(signal_bundle_builder):
    signal_bundle_builder.version = "1"
    empty_build_bundle = signal_bundle_builder.build()
    assert empty_build_bundle.identifier == "hello builder identifier"
    assert empty_build_bundle.signals == []
    assert empty_build_bundle.version == "1"
    signal_bundle_builder.register_signal("plop_topic", {"content_key1": 0}, other_kwarg=11)
    full_build_bundle = signal_bundle_builder.build()
    assert full_build_bundle.identifier == "hello builder identifier"
    assert full_build_bundle.signals is signal_bundle_builder.signals
    assert full_build_bundle.version == "1"


def test_reset(signal_bundle_builder):
    assert signal_bundle_builder.signals == []
    signal_bundle_builder.register_signal("plop_topic", {"content_key1": 0}, other_kwarg=11)
    assert len(signal_bundle_builder.signals) == 1
    signal_bundle_builder.reset()
    assert signal_bundle_builder.signals == []



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import os
import mock

import octobot_commons.pretty_printer as pretty_printer
import octobot_commons.constants as constants


def test_get_min_string_from_number():
    assert pretty_printer.get_min_string_from_number(1) == "1"
    assert pretty_printer.get_min_string_from_number(-1) == "-1"
    assert pretty_printer.get_min_string_from_number(1.000000000001) == "1"
    assert pretty_printer.get_min_string_from_number(1.000000000001, max_digits=1) == "1"
    assert pretty_printer.get_min_string_from_number(1.00000001, max_digits=8) == "1.00000001"
    assert pretty_printer.get_min_string_from_number(0.00000001, max_digits=8) == "0.00000001"
    assert pretty_printer.get_min_string_from_number(-0.000000009, max_digits=8) == "-0.00000001"
    assert pretty_printer.get_min_string_from_number(100.00000001, max_digits=8) == "100.00000001"
    assert pretty_printer.get_min_string_from_number(-100.00000001, max_digits=8) == "-100.00000001"
    assert pretty_printer.get_min_string_from_number(100.09, max_digits=1) == "100.1"
    assert pretty_printer.get_min_string_from_number(100.06, max_digits=1) == "100.1"
    assert pretty_printer.get_min_string_from_number(100.05, max_digits=1) == "100"
    assert pretty_printer.get_min_string_from_number(100.04, max_digits=1) == "100"
    assert pretty_printer.get_min_string_from_number(101.04, max_digits=1) == "101"
    assert pretty_printer.get_min_string_from_number(100.00000000, max_digits=8) == "100"
    assert pretty_printer.get_min_string_from_number(-100.00000000, max_digits=8) == "-100"
    # with computed max_digits
    assert pretty_printer.get_min_string_from_number(-100.00000001) == "-100"
    assert pretty_printer.get_min_string_from_number(100.00000001) == "100"
    assert pretty_printer.get_min_string_from_number(-1.12345678) == "-1.12"
    assert pretty_printer.get_min_string_from_number(1.12345678) == "1.12"
    assert pretty_printer.get_min_string_from_number(0.12345678) == "0.1235"
    assert pretty_printer.get_min_string_from_number(-0.12345678) == "-0.1235"
    assert pretty_printer.get_min_string_from_number(0.0045678) == "0.004568"
    assert pretty_printer.get_min_string_from_number(-0.0045678) == "-0.004568"
    assert pretty_printer.get_min_string_from_number(0.0000456789) == "0.00004568"
    assert pretty_printer.get_min_string_from_number(-0.0000456789) == "-0.00004568"


def test_round_with_decimal_count():
    assert pretty_printer.round_with_decimal_count(None) == 0
    assert pretty_printer.round_with_decimal_count(1) == 1
    assert pretty_printer.round_with_decimal_count(-1) == -1
    assert pretty_printer.round_with_decimal_count(1.000000000001, max_digits=1) == 1
    if not os.getenv('CYTHON_IGNORE'):
        with mock.patch.object(pretty_printer, "get_min_string_from_number", mock.Mock(return_value="1")) \
                as get_min_string_from_number_mock:
            pretty_printer.round_with_decimal_count(None, max_digits=4)
            get_min_string_from_number_mock.assert_not_called()
            pretty_printer.round_with_decimal_count(1.011, max_digits=4)
            get_min_string_from_number_mock.assert_called_once_with(1.011, 4)


def test_global_portfolio_pretty_print():
    portfolio = {
        "BTC": {constants.PORTFOLIO_TOTAL: decimal.Decimal(1)},
        "ETH": {constants.PORTFOLIO_TOTAL: decimal.Decimal(2)},
        "PLOP": {constants.PORTFOLIO_TOTAL: decimal.Decimal("0.444")},
        "ADA": {constants.PORTFOLIO_TOTAL: decimal.Decimal(0)}
    }
    # without ref market
    for res in (
        pretty_printer.global_portfolio_pretty_print(portfolio),
        pretty_printer.global_portfolio_pretty_print(portfolio, markdown=True)
    ):
        assert "BTC" in res
        assert "1" in res
        assert "ETH" in res
        assert "2" in res
        assert "PLOP" in res
        assert "0.444" in res
        assert "ADA" not in res
    # with ref market
    for res in (
        pretty_printer.global_portfolio_pretty_print(portfolio, ref_market_name="BTC"),
        pretty_printer.global_portfolio_pretty_print(portfolio, ref_market_name="BTC", markdown=True)
    ):
        assert "BTC" in res
        assert "1" in res
        assert "ETH" in res
        assert "2" in res
        assert "PLOP" in res
        assert "0.444" in res
        assert "ADA" not in res
    currency_values = {"ETH": decimal.Decimal("1.5")}
    for res in (
        pretty_printer.global_portfolio_pretty_print(portfolio, currency_values=currency_values,
                                                     ref_market_name="BTC"),
        pretty_printer.global_portfolio_pretty_print(portfolio, currency_values=currency_values,
                                                     ref_market_name="BTC", markdown=True)
    ):
        assert "BTC" in res
        assert "1" in res
        assert "ETH" in res
        assert "2" in res
        assert "3" in res
        assert "PLOP" in res
        assert "0.444" in res
        assert "ADA" not in res
    # with separator
    for res in (
        pretty_printer.global_portfolio_pretty_print(portfolio, currency_values=currency_values,
                                                     ref_market_name="BTC", separator="111"),
        pretty_printer.global_portfolio_pretty_print(portfolio, currency_values=currency_values,
                                                     ref_market_name="BTC", markdown=True, separator="111")
    ):
        assert "BTC" in res
        assert "1" in res
        assert "ETH" in res
        assert "2" in res
        assert "3" in res
        assert "PLOP" in res
        assert "0.444" in res
        assert "ADA" not in res
        assert "111" in res



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import numpy as np


def normalize_data(data):
    """
    Normalize the specified data
    :param data: the data to normalize
    :return: normalized data
    """
    if data.size > 1:
        return (data - np.mean(data)) / (data.max() - data.min())
    return data


def drop_nan(data):
    """
    Drop nan of a numpy array
    :param data: the numpy array
    :return: the numpy array without nan value
    """
    return data[~np.isnan(data)]


def mean(number_list):
    """
    Return the list average
    :param number_list: the list to use
    :return: the list average
    """
    return sum(number_list) / len(number_list) if number_list else 0


def shift_value_array(array, shift_count=-1, fill_value=np.nan, dtype=np.float64):
    """
    Shift a numpy array
    :param array: the numpy array
    :param shift_count: the shift direction (- / +) and counter
    :param fill_value: the new value of the shifted indexes
    :param dtype: the type of the numpy array (and also the :fill_value:)
    :return: the shifted array
    """
    new_array = np.empty_like(array, dtype=dtype)
    if shift_count > 0:
        new_array[:shift_count] = fill_value
        new_array[shift_count:] = array[:-shift_count]
    elif shift_count < 0:
        new_array[shift_count:] = fill_value
        new_array[:shift_count] = array[-shift_count:]
    else:
        new_array[:] = array
    return new_array
